import os

from collections.abc import Iterator
from contextlib import suppress
from pathlib import Path
from tempfile import TemporaryFile
from typing import Optional

from hishel import CacheClient, FileStorage
from httpx import Client, URL  # noqa: F401

from .config import read as read_config
from .__version__ import __version__

__url__ = "https://github.com/gbenson/hive"


def user_cache_path() -> Optional[Path]:
    """https://pkg.go.dev/os#UserCacheDir
    """
    dirname = os.environ.get("XDG_CACHE_HOME")
    if dirname:
        return Path(dirname)
    with suppress(RuntimeError):
        return Path.home() / ".cache"
    return None


def _default_cache_path_options() -> Iterator[Path]:
    if (path := user_cache_path()):
        yield path
    for dirname in ("/var/cache", "/var/tmp", "/tmp"):
        yield Path(dirname)


def _special_mkdir(path: Path, mode: int = 0o700) -> Path:
    path = path.resolve()
    path.mkdir(mode=mode, parents=True, exist_ok=True)
    path.chmod(mode)
    with TemporaryFile(dir=path):
        pass
    return path


def _default_cache_path() -> Optional[Path]:
    for basedir in _default_cache_path_options():
        with suppress(PermissionError):
            return _special_mkdir(basedir / "hive" / "hishel")
    return None


DEFAULT_CACHE_PATH = _default_cache_path()


def _default_user_agent(name: str = "HiveBot") -> str:
    config_key = name.lower()
    try:
        template = read_config(config_key)[config_key]["user_agent"]
    except KeyError:
        template = f"{name}/{{version}} (bot; +{__url__})"
    return template.format(version=__version__)


DEFAULT_USER_AGENT = _default_user_agent()


DEFAULT_CLIENT: Client = CacheClient(
    http2=True,
    storage=FileStorage(base_path=DEFAULT_CACHE_PATH),
)
DEFAULT_CLIENT.headers["User-Agent"] = DEFAULT_USER_AGENT


globals().update(
    (attr, getattr(DEFAULT_CLIENT, attr))
    for attr in (
            "delete",
            "get",
            "head",
            "options",
            "patch",
            "post",
            "put",
            "request",
            "stream",
    )
)



from itertools import count
from typing import Callable, Optional


def once(func):
    """Decorator that ensures func runs only once.
    """
    counter = count()
    def wrapper(*args, **kwargs):
        if next(counter) != 0:
            return None
        return func(*args, **kwargs)
    return wrapper


def chained(*funcs) -> Optional[Callable]:
    """Return a callable that's equivalent to calling each not-None
    supplied callable in sequence with the same arguments.  If all
    supplied callables are None then None is returned.
    """
    funcs = [func for func in funcs if func]
    if not funcs:
        return None
    if len(funcs) == 1:
        return funcs[0]

    def result(*args, **kwargs):
        for func in funcs:
            func(*args, **kwargs)

    return result



from . import read_config

try:
    from openai import (
        APIStatusError,
        OpenAI as _OpenAI,
    )

except ModuleNotFoundError:
    class APIStatusError(Exception):
        pass

    class _OpenAI:
        def __init__(self, *args, **kwargs):
            raise ModuleNotFoundError("hive-common[openai] not installed")


class OpenAI(_OpenAI):
    def __init__(self, **kwargs):
        api_key = kwargs.pop("api_key", None)
        if not api_key:
            api_key = read_config("openai")["api_key"]
        super().__init__(api_key=api_key, **kwargs)



import pytest

from hive.common.testing import assert_is_valid_version


@pytest.mark.parametrize(
    "invalid_version",
    (("1.2-3"),
     ("1.2-3-final"),
     ))
def test_is_valid_version(invalid_version):
    with pytest.raises(AssertionError):
        assert_is_valid_version(invalid_version)



from hive.common import read_resource


def test_read_resource():
    with open(__file__) as fp:
        expected = fp.read()
    result = read_resource("test_read_resource.py")
    assert result == expected



import inspect
import os


def read_resource(filename: str, mode: str = "r") -> str | bytes:
    if not os.path.isabs(filename):
        caller_filename = inspect.currentframe().f_back.f_code.co_filename
        filename = os.path.join(os.path.dirname(caller_filename), filename)
    with open(filename, mode) as fp:
        return fp.read()



import logging

import pytest

from hive.common import ArgumentParser
from hive.common.testing import want_to_see


def test_hive_default_log_level(monkeypatch):
    logging_basicconfig_calls = []

    def mock_bc(*args, **kwargs):
        nonlocal logging_basicconfig_calls
        logging_basicconfig_calls.append((args, kwargs))

    with monkeypatch.context() as mp:
        mp.setattr(logging, "basicConfig", mock_bc)
        mp.delenv("LL", raising=False)
        _ = ArgumentParser().parse_args()

    assert logging_basicconfig_calls == [((), {"level": logging.INFO})]


def test_library_default_log_level(monkeypatch):
    def mock_bc(*args, **kwargs):
        pytest.fail("logging.basicConfig() was called")

    with monkeypatch.context() as mp:
        mp.setattr(logging, "basicConfig", mock_bc)
        mp.delenv("LL", raising=False)
        parser = ArgumentParser()
        parser.DEFAULT_LOGLEVEL = None
        _ = parser.parse_args()


@pytest.mark.parametrize(
    ("user_level", "expect_level"),
    (("debug", logging.DEBUG),
     ("DEBUG", logging.DEBUG),
     ("10", logging.DEBUG),
     ("11", 11),
     ("WARNING", logging.WARNING),
     ("critical", logging.CRITICAL),
     ))
def test_getenv_log_level(user_level, expect_level, monkeypatch):
    loggers = []

    def logger_name(i):
        return f"{__name__}.test_getenv_log_level.{i}"

    def mock_bc(level):
        logger = logging.getLogger(logger_name(len(loggers)))
        logger.setLevel(level)
        loggers.append(logger)

    with monkeypatch.context() as mp:
        mp.setattr(logging, "basicConfig", mock_bc)
        mp.setenv("LL", user_level)
        _ = ArgumentParser().parse_args()

    assert len(loggers) == 1
    logger = loggers[0]
    assert logger.name == logger_name(0)
    assert logger.level == expect_level


def test_unknown_log_level(monkeypatch, caplog):
    def mock_bc(level):
        logger = logging.getLogger(f"{__name__}.test_unknown_log_level")
        logger.setLevel(level)
        pytest.fail(f"logger.setLevel({level!r}) didn't raise")

    with monkeypatch.context() as mp:
        mp.setattr(logging, "basicConfig", mock_bc)
        mp.setenv("LL", "gonk")

        with caplog.at_level(logging.WARNING):
            _ = ArgumentParser().parse_args()

    want_to_see(caplog, "Ignoring LL='GONK'")



from hive.common.llm import LLMToolbox, llm_tool


class TestToolbox(LLMToolbox):
    @llm_tool
    def get_current_weather(self, city: str):
        """Get the current weather for a city.

        :param city: The name of the city.
        """

    def test_llm_tool_decorated(self):
        assert self.get_current_weather.__llm_tool__

    def test_llm_tools_property(self):
        assert self.llm_tools == [
            {
                "type": "function",
                "function": {
                    "name": "get_current_weather",
                    "description": "Get the current weather for a city.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "city": {
                                "type": "string",
                                "description": "The name of the city.",
                            },
                        },
                        "required": ["city"],
                    },
                },
            },
        ]



import pytest

try:
    from pep440_version_utils import Version
except ImportError:  # pragma: no cover
    pass

from .config import DEFAULT_READER


def assert_is_valid_version(version: str):
    try:
        assert str(Version(version)) == version
    except Exception as e:
        raise AssertionError(f"Invalid version {version!r}") from e


@pytest.fixture
def test_config_dir(tmp_path, monkeypatch):
    dirname = str(tmp_path)
    with monkeypatch.context() as m:
        m.setattr(DEFAULT_READER, "search_path", [dirname])
        yield dirname


def want_to_see(caplog, msg):
    failure_detail = f"didn't see: {msg}"
    assert any(r.getMessage() == msg for r in caplog.records), failure_detail



from collections.abc import Iterable
from dataclasses import dataclass
from inspect import get_annotations, stack
from typing import Any, Callable, ClassVar, TypeVar


class LLMToolbox:
    @property
    def llm_tools(self) -> list[dict[str, Any]]:
        locs = [(frame.filename, frame.lineno) for frame in stack()]
        if locs[1] in locs[2:]:
            return []  # recursion!
        return list(self._llm_tools)

    @property
    def _llm_tools(self) -> Iterable[dict[str, Any]]:
        for name in dir(self):
            if name.startswith("_"):
                continue
            candidate = getattr(self, name)
            if not getattr(candidate, "__llm_tool__", False):
                continue
            yield LLMTool(name, candidate).definition


@dataclass
class LLMTool:
    name: str
    func: Callable

    @property
    def definition(self) -> dict[str, Any]:
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": self.parameters,
            },
        }

    @property
    def description(self) -> str:
        return self.func.__doc__.partition(":param ")[0].strip()

    @property
    def parameters(self) -> dict[str, Any]:
        params = dict(self._parameters)
        return {
            "type": "object",
            "properties": params,
            "required": list(params.keys()),
        }

    @property
    def _parameters(self) -> Iterable[str, dict[str, Any]]:
        for param_doc in self.func.__doc__.split(":param ")[1:]:
            name, description = param_doc.split(":", 1)
            _type = get_annotations(self.func)[name]
            yield name, {
                "type": self._TYPE_NAMES[_type],
                "description": description.strip(),
            }

    _TYPE_NAMES: ClassVar[dict[TypeVar, str]] = {
        int: "integer",
        str: "string",
    }
    _TYPE_NAMES.update(
        (key.__name__, value)
        for key, value in list(_TYPE_NAMES.items())
    )


def llm_tool(func):
    """Decorator."""
    func.__llm_tool__ = True
    return func



from .argument_parser import HiveArgumentParser as ArgumentParser
from .buffers import SmallCircularBuffer
from .config import read as read_config
from .resource import read_resource
from .uuid import parse_uuid



import os

from hive.common.openai import _OpenAI, OpenAI
from hive.common.testing import test_config_dir  # noqa: F401


def test_openai(test_config_dir, monkeypatch):  # noqa: F811
    with open(os.path.join(test_config_dir, "openai.env"), "w") as fp:
        print("OPENAI_API_KEY=BLrnKp+hf5Qyn7UqH2RFHGX/smwlqDW", file=fp)

    calls = []
    def log_call(*args, **kwargs):
        calls.append((args, kwargs))

    with monkeypatch.context() as mp:
        mp.setattr(_OpenAI, "__init__", log_call)
        client = OpenAI()

    assert calls == [
        ((client,), {"api_key": "BLrnKp+hf5Qyn7UqH2RFHGX/smwlqDW"}),
    ]



import json

from http.server import BaseHTTPRequestHandler, HTTPServer

import pytest

from hive.common import httpx
from hive.common.socketserver import serving


class TestServerRequestHandler(BaseHTTPRequestHandler):
    __test__ = False

    def do_GET(self):
        response = json.dumps({
            "requestline": self.requestline,
            "headers": dict(
                (key.lower(), value)
                for key, value in self.headers.items()
            ),
        }).encode("utf-8")
        self.send_response(200)
        self.send_header("Content-type", "application/json")
        self.send_header("Content-Length", str(len(response)))
        self.end_headers()
        self.wfile.write(response)


@pytest.fixture
def test_server():
    server = HTTPServer(("127.0.0.1", 0), TestServerRequestHandler)
    with serving(server):
        host, port = server.server_address
        server.base_url = f"http://{host}:{port}"
        yield server


def test_get(test_server):
    r = httpx.get(f"{test_server.base_url}/hello-world")
    r.raise_for_status()
    req = r.json()
    assert req["requestline"].startswith("GET /hello-world HTTP/1.")
    assert req["headers"]["host"] == ("%s:%d" % test_server.server_address)
    user_agent = req["headers"]["user-agent"]
    assert user_agent.startswith("HiveBot/0.0.")
    assert " (bot; +https://" in user_agent



import json
import os
import yaml

from typing import Optional
from collections.abc import Iterable


def user_config_dir() -> Optional[str]:
    """https://pkg.go.dev/os#UserConfigDir"""
    dirname = os.environ.get("XDG_CONFIG_HOME")
    if dirname:
        return dirname
    homedir = os.environ.get("HOME")
    if homedir:
        return os.path.join(homedir, ".config")
    return None


class Reader:
    def __init__(self, subdirs: Iterable[str] = ("hive",)):
        self.search_path = [
            os.path.join(dirname, *subdirs)
            for dirname in (user_config_dir(), "/etc")
            if dirname
        ]
        self.search_path.append("/run/secrets")
        self.search_exts = [
            "",
            ".yml",
            ".yaml",
            ".json",
            ".env",
        ]

    def get_filename_for(self, key: str) -> str:
        for dirname in self.search_path:
            basename = os.path.join(dirname, key)
            for ext in self.search_exts:
                filename = basename + ext
                if os.path.isfile(filename):
                    return filename
        raise KeyError(key)

    def read(self, key: str, type: str = "yaml"):
        filename = self.get_filename_for(key)
        ext = os.path.splitext(filename)[1].lstrip(".")
        if ext in {"env", "json"}:
            type = ext
        return getattr(self, f"_read_{type}")(filename)

    def _read_env(self, filename):
        with open(filename) as fp:
            lines = fp.readlines()
        lines = [line.split("#", 1)[0].strip() for line in lines]
        items = [line.split("=", 1) for line in lines if line]
        result = dict((k.rstrip(), v.lstrip()) for k, v in items)
        prefix = os.path.splitext(
            os.path.basename(filename))[0].upper().replace("-", "_") + "_"
        extras = dict(
            (key[len(prefix):].lower(), value)
            for key, value in result.items()
            if key.startswith(prefix)
        )
        extras.update(result)
        return extras

    def _read_json(self, filename):
        with open(filename) as fp:
            return json.load(fp)

    def _read_yaml(self, filename):
        with open(filename) as fp:
            return yaml.safe_load(fp)


DEFAULT_READER = Reader()

read = DEFAULT_READER.read



import logging

from contextlib import contextmanager
from socketserver import BaseServer
from threading import Thread

from .units import SECONDS

logger = logging.getLogger(__name__)


@contextmanager
def serving(
        server: BaseServer,
        *,
        shutdown_timeout: float = 30 * SECONDS,
        daemon: bool = True,
) -> Thread:
    """Run a :class:`socketserver.BaseServer` in another thread.
    """
    thread = Thread(target=server.serve_forever, daemon=daemon)
    logger.info("%s:%s: Starting server", thread, server)
    thread.start()
    try:
        logger.info("%s:%s: Server started", thread, server)
        yield thread
    finally:
        logger.info("%s: %s: Stopping server", thread, server)
        server.shutdown()
        logger.debug("%s: Waiting for thread exit", thread)
        thread.join(timeout=30 * SECONDS)
        logger.info("%s: %s: Server stopped", thread, server)



BYTE = BYTES = 1
KiB = BYTE << 10
MiB = KiB << 10
GiB = MiB << 10
TiB = GiB << 10

SECOND = SECONDS = 1
MINUTE = MINUTES = 60 * SECONDS
HOUR = HOURS = 60 * MINUTES
DAY = DAYS = 24 * HOURS  # usually

MILLISECOND = MILLISECONDS = SECOND / 1000
MICROSECOND = MICROSECONDS = MILLISECOND / 1000



import logging

from argparse import ArgumentParser

from .logging import getenv_log_level

logger = logging.getLogger(__name__)


class HiveArgumentParser(ArgumentParser):
    DEFAULT_EPILOG = "Run with LL=debug for so much extra logging."
    DEFAULT_LOGLEVEL = logging.INFO

    def parse_args(self, *args, **kwargs):
        if not self.epilog:
            self.epilog = self.DEFAULT_EPILOG

        if (log_level := getenv_log_level(default=self.DEFAULT_LOGLEVEL)):
            try:
                logging.basicConfig(level=log_level)
            except ValueError:
                logger.warning(f"Ignoring LL={log_level!r}")
        return super().parse_args(*args, *kwargs)



from uuid import RFC_4122, UUID


def parse_uuid(uuid: str | UUID) -> UUID:
    if not isinstance(uuid, UUID):
        uuid = UUID(uuid)

    if uuid.variant != RFC_4122:
        raise ValueError(uuid)
    if uuid.version != 4:
        raise ValueError(uuid)

    return uuid



import os

from typing import Optional


def getenv_log_level(default=None) -> Optional[int | str]:
    level = os.environ.get("LL", default)
    if not level:
        return None
    try:
        return int(level)
    except ValueError:
        return level.upper()



class SmallCircularBuffer:
    def __init__(self, num_items, coerce=None):
        sentinel = object()
        if not coerce:
            coerce = lambda x: x  # noqa: E731
        self._coerce = coerce
        self._items = [sentinel] * num_items
        self._index = 0

    def __contains__(self, item):
        return self._coerce(item) in self._items

    def add(self, item):
        self._items[self._index] = self._coerce(item)
        self._index = (self._index + 1) % len(self._items)
